from functools import reduce
from typing import Callable, Tuple, TypeVar
from .helpers import flatten
from .types import PypelyTuple

T = TypeVar("T")

def pipeline(*funcs: Callable) -> Callable:
    def _reducer(func1, func2):
        return lambda *x: func2(func1(*x))

    return reduce(_reducer, funcs)


def fork(*funcs: Callable) -> Callable[..., Tuple]:
    return lambda *x: PypelyTuple(*[func(*x) for func in funcs])


def to(obj: T, *set_fields) -> Callable[..., T]:
    def _inner(vals: PypelyTuple) -> T:
        vals_flattened = flatten(vals)
        if not set_fields == ():
            assert len(vals_flattened) == len(set_fields)
            fields_named = {field_name: val for field_name, val in zip(set_fields, vals_flattened)}
            return obj(**fields_named)
        else:
            return obj(*vals_flattened)
    
    return _inner


def merge(func: Callable[..., T]) -> Callable[[Tuple], T]:
    return lambda branches: func(*flatten(branches))


def identity(*x):
    if len(x) == 1:
        return x[0]
    return x



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

__version__ = "0.1.2"

setuptools.setup(
    name="pypely",
    version=__version__,
    description="Make data processing easy",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/stoney95/pypely",
    author="Stoney95",
    author_email="simon@steinheber.info",
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages(where="src"),
    package_dir={'': 'src'},
    python_requires=">=3.7",
    install_requires=[],
    extras_require={
        'dev': []
    },
)


from typing import Callable
from inspect import getfullargspec
from .types import PypelyTuple


def reduce_by(func: Callable) -> Callable:
    num_args = len(getfullargspec(func).args)
    return lambda *x: flatten(PypelyTuple(func(*x[:num_args]), *x[num_args:]))


def flatten(_tuple: PypelyTuple) -> PypelyTuple:
    result = []
    if isinstance(_tuple, PypelyTuple):
        for elem in _tuple:
            if isinstance(elem, PypelyTuple):
                if any(isinstance(x, PypelyTuple) for x in elem):
                    result += list(flatten(elem))
                else:
                    result += list(elem)
            else:
                result.append(elem)
        return PypelyTuple(*result)
    raise ValueError(f"You can use flatten only with 'PypelyTuple'. Input is of type: {type(_tuple)}")


def side_effect(func: Callable):
    def __run_func(*_input):
        func(*_input)

        # TODO: find better solution
        if len(_input) == 1:
            return _input[0]
        return _input

    return lambda *x: __run_func(*x)


def optional(func: Callable, cond: bool):
    return lambda *x:  func(*x) if cond else x


head = lambda x: x[0]
last = lambda x: x[-1]
rest = lambda x: PypelyTuple(*x[1:])